#!/usr/bin/env python

'''
Web Log Cleaning class.

Contents:
- Cleaner: logfile cleaning class
  clean_log = Cleaner(log object)
  methods:
	- clean_log.getlogent()
  variables:
	- clean_log.directory_index  [ list of site URL's to treat as local ]
	- clean_log.cache_size  [ maximum size of url cache ]

- test: test function

'''

# (c) 1998 Copyright Mark Nottingham
# <mnot@pobox.com>
#
# This software may be freely distributed, modified and used,
# provided that this copyright notice remain intact.
#
# This software is provided 'as is' without warranty of any kind.


# Web Log Cleaning
#
# The log object fed to a Cleaner MUST be preprocessed with a WebLogUrlParser.
#
# The Cleaner class will normalise the log entries as much as possible, 
# resulting in more accurate statistics for analysis. Although its efforts
# are not perfect (particularly with path collapsing, due to the differing
# natures of browsers and servers), it tries to 'do the right thing'.
# 
# This class will do the following to a log fed to it, for both request URL
# and referer (if available):
# 
# - collapse out ./ ../ // and other sequences where appropriate
# 
# - replace %7e and %7E with ~
# 
# Additionally, 
# 
# - if a request URL is for a page specified in the directory_index list,
#   that page is trimmed from the path
# 
# - referer hosts are normailsed to lowercase
# 
# - referer hosts that specify a port which is the default port for that
#   service will have it removed
#
# - referer news: URL's that specify a host have it removed


__version__ = '0.92'



from string import join, rfind, lower
from regsub import gsub
from urlparse import urlunparse
import regex

_replace_pat = regex.compile('\(/[^/]+/\.\./?\|/\./\|//\|/\.$\|/\.\.$\)')
_user_pat = regex.compile('%7e', regex.casefold)

_default_port = {	'http': ':80',
					'https': ':443',
					'gopher': ':70',
					'news': ':119',
					'snews': ':563',
					'nntp': ':119',
					'snntp': ':563',
					'ftp': ':21',
					'telnet': ':23',
					'prospero': ':191',
				}

_relative_scheme = {	'http': 1,
						'https': 1,
						'news': 1,
						'snews': 1,
						'nntp': 1,
						'snntp': 1,
						'ftp': 1,
						'file': 1,
						'': 1
					}


class Cleaner:
	def __init__(self, log):
		self.log = log
		self.directory_index = []
		self.set_url = 0
		self.set_ref = 0
		self.cache_size = 1000
		self._cache = {'url': {}, 'ref': {}}


	def __getattr__(self, attr):
		try:
			return getattr(self.log, attr)
		except AttributeError:
			raise AttributeError, attr


	def __setattr__(self, attr, value):
		if attr == 'directory_index':
			self._dir_pat = "/\(" + join(value, "\|") + "\)$"
			self._dir_comp = regex.compile(self._dir_pat)
		self.__dict__[attr] = value


	def getlogent(self):

		''' Increment position in the log and populate requested attributes '''

		if self.log.getlogent():
			self.__dict__['url'] = self._clean(self.log.url, 'url')
			try:
				self.__dict__['referer'] = self._clean(self.log.referer, 'ref')
			except AttributeError:
				pass
			return 1
		else:
			return 0


	def _clean(self, url, url_type):

		if len(self._cache[url_type]) > self.cache_size:
			self._cache[url_type] = {}
			
		try:
			_cached = self._cache[url_type][url]
		except KeyError:
			_scheme = getattr(self.log, url_type + "_scheme")
			_path = getattr(self.log, url_type + "_path")

			if _relative_scheme.get(_scheme, 0):
				_last_path = _path
				while 1:
					_path = gsub(_replace_pat, '/', _path)
					if _last_path == _path:
						break
					_last_path = _path

			_path = gsub(_user_pat, '~', _path)

			if url_type == 'url':		 	
				_scheme, _host = '', ''
				if self._dir_comp:
					if self._dir_comp.search(_path) > -1:
						_slash_index = rfind(_path, '/') + 1
						_path = _path[:_slash_index]

			elif url_type == 'ref':
				_host = lower(self.log.ref_host)
				_colon_index = rfind(_host, ':')
				if _colon_index:
					if _host[_colon_index:] == _default_port.get(_scheme, '#'):
						_host = _host[:_colon_index]
				if _scheme == 'news':
					_slash_index = rfind(_path, '/') + 1
					_path = _path[_slash_index:]
				
			self._cache[url_type][url] = (	_scheme, 
											_host, 
											_path, 
											getattr(self, url_type + "_parameters"),
											getattr(self, url_type + "_query"),
											getattr(self, url_type + "_fragment")
										)
			_cached = self._cache[url_type][url]


		self.__dict__[url_type + "_scheme"] = _cached[0]
		self.__dict__[url_type + "_host"] = _cached[1]
		self.__dict__[url_type + "_path"] = _cached[2]
		return(urlunparse(_cached))
	



def test():

	''' basic test suite- modify at will to test full functionality '''

	import sys, CombinedWebLog, WebLogUrlParse

	file = sys.stdin
	log = CombinedWebLog.Parser(file)
	p_log = WebLogUrlParse.UrlParser(log)

	clean_log = Cleaner(p_log)
	clean_log.directory_index = ['index.html', 'index.htm']

	while clean_log.getlogent():
		print "%s\n%s %s" % (log.url, clean_log.url, clean_log.url_path)
		print "%s\n%s %s %s" % (log.referer, clean_log.referer, clean_log.ref_host, clean_log.ref_path)
		print


def test_clean():

	''' function to test _clean's operation '''
	
	from urlparse import urlparse

	cleaner = Cleaner('')
	cleaner.directory_index = ['index.html', 'index.htm']
	

	u_tests = { '/foo/../bar':				'/bar',
				'/foo/index.html':			'/foo/',
				'/foo/index.html/':			'/foo/index.html/',
				'/index.html.':				'/index.html.',
				'/index.html':				'/',
				'/index.htm':				'/',
				'/index.ht':				'/index.ht'
			}

	r_tests = {	'/foo/bar/.':				'/foo/bar/', 
				'/foo/bar/./':				'/foo/bar/',
				'/foo/bar/..':				'/foo/',
				'/foo/bar/../': 			'/foo/',
				'/foo/bar/../baz': 			'/foo/baz',
				'/foo/bar/../..': 			'/',
				'/foo/bar/../../': 			'/',
				'/foo/bar/../../baz': 		'/baz',
				'/foo/bar/../../../baz':	'/../baz',
				'/foo/bar/../../../../baz':	'/baz',
				'/./foo':					'/foo',
				'/../foo':					'/../foo',
				'/foo.':					'/foo.',
				'/.foo':					'/.foo',
				'/foo..':					'/foo..',
				'/..foo':					'/..foo',
				'/./../foo':				'/foo',
				'/./foo/.':					'/foo/',
				'/foo/./bar':				'/foo/bar',
				'/foo/../bar':				'/bar',
				'/foo//':					'/foo/',
				'/foo///bar//':				'/foo/bar/',	
				'news:alt.this.group':		'news:alt.this.group',
				'news://foo.com/user-full.article.number@whatever.server.net':	'news:user-full.article.number@whatever.server.net',
				'news:user-full.article.number@whatever.server.net':	'news:user-full.article.number@whatever.server.net',
				'http://www.foo.com:80/foo':	'http://www.foo.com/foo',
				'http://www.foo.com:8000/foo':	'http://www.foo.com:8000/foo',
				'-':						'-',
			}

	n_correct, n_fail = 0, 0
	test_types = { 	'url': u_tests,
					'ref': r_tests,
				}
	type_keys = test_types.keys()
	
	for ty in type_keys:
		test_keys = test_types[ty].keys()
		test_keys.sort()		
	
		for i in test_keys:
			print 'ORIGINAL:', i
			(	cleaner.__dict__[ty + '_scheme'],
				cleaner.__dict__[ty + '_host'],
				cleaner.__dict__[ty + '_path'],
				cleaner.__dict__[ty + '_parameters'],
				cleaner.__dict__[ty + '_query'],
				cleaner.__dict__[ty + '_fragment']	) = urlparse(i)
			
			cleaned = cleaner._clean(i, ty)
			answer = test_types[ty][i]
			print ty, 'CLEANED: ', cleaned
			print ty, 'CORRECT: ', answer
			if cleaned != answer:
				print "*** TEST FAILED"
				n_fail = n_fail + 1
			else:
				n_correct = n_correct + 1
			print
		
	print "TOTAL CORRECT:", n_correct
	print "TOTAL FAILURE:", n_fail




if __name__ == '__main__':
	test()
